@@ -3,14 +3,17 @@
 from __future__ import division
 import re
 import sys
-import textwrap
 
 from pyAMI.objects import DatasetInfo, RunPeriod
 from pyAMI.schema import *
+from pyAMI.utils import *
 from pyAMI.defaults import YEAR, STREAM, TYPE, PROJECT, PRODSTEP
 
 
-DATA_PATTERN = re.compile('^(?P<project>\w+).(?P<run>[0-9]+).(?P<stream>[a-zA-Z_\-0-9]+).(recon|merge).(?P<type>[a-zA-Z_\-0-9]+).(?P<version>\w+)$')
+DATA_PATTERN = re.compile(
+    '^(?P<project>\w+).(?P<run>[0-9]+).'
+    '(?P<stream>[a-zA-Z_\-0-9]+).(recon|merge).'
+    '(?P<type>[a-zA-Z_\-0-9]+).(?P<version>\w+)$')
 
 ESD_VERSION_PATTERN = '(?P<la>f|r)(?P<lb>[0-9]+)'
 AOD_VERSION_PATTERN = ESD_VERSION_PATTERN + '_(?P<ma>m|p)(?P<mb>[0-9]+)'
@@ -21,7 +24,7 @@ AOD_VERSION_PATTERN = re.compile('^%s$' % AOD_VERSION_PATTERN)
 NTUP_VERSION_PATTERN = re.compile('^%s$' % NTUP_VERSION_PATTERN)
 
 
-def __clean_dataset(dataset):
+def _clean_dataset(dataset):
     """
     Remove trailing slashes
 
@@ -33,97 +36,6 @@ def __clean_dataset(dataset):
     return dataset.rstrip('/')
 
 
-def flatten_results(things, fields):
-    """
-    Convert list of dicts to list of tuples
-    tuple elements are in the same order as in fields
-
-    *things*: list
-        list of dicts
-
-    *fields*: [ tuple | list ]
-        keys of each dict to include in tuples
-    """
-    return [[thing[field.split('.')[-1]] for field in fields] for thing in things]
-
-
-def parse_fields(fields, table):
-    """
-    *fields*: [tuple | list | str ]
-        field names. If a string, then comma separated
-
-    *table*: pyAMI.ami.schema.Table
-        the schema table containing these fields
-    """
-    query_fields = []
-    if fields:
-        if isinstance(fields, basestring):
-            fields = fields.split(',')
-        for name in fields:
-            name = validate_field(name, table)
-            query_fields.append(name)
-    return query_fields
-
-
-def validate_field(field, table):
-
-    try:
-        if field in table.fields.values():
-            return field
-        name = field.replace('-', '_')
-        if '.' in name:
-            foreign_name, foreign_field = name.split('.')
-            try:
-                foreign_entity = table.foreign[foreign_name]
-            except KeyError, AttributeError:
-                raise ValueError('%s is not associated with %s' % (table.__name__, foreign_name))
-            if foreign_field not in foreign_entity.fields.values():
-                foreign_field = foreign_entity.fields[foreign_field]
-            name = '%s.%s' % (foreign_name, foreign_field)
-        else:
-            name = table.fields[name]
-    except KeyError:
-        raise ValueError(('field %s does not exist\n' % name) + \
-                         'valid fields are:\n\t' + '\n\t'.join(table.fields.keys()))
-    return name
-
-
-def print_table(table, sep='  ',
-                wrap_last=False,
-                wrap_width=50,
-                header=False,
-                vsep=None,
-                stream=None):
-
-    if stream is None:
-        stream = sys.stdout
-    # reorganize data by columns
-    cols = zip(*table)
-    # compute column widths by taking maximum length of values per column
-    col_widths = [max(len(value) for value in col) for col in cols]
-    # create a suitable format string
-    format = sep.join(['%%-%ds' % width for width in col_widths[:-1] ] + ['%s'])
-    # print each row using the computed format
-    if wrap_last:
-        total_width = sum(col_widths[:-1]) + (len(sep) * (len(cols) - 1)) + wrap_width
-        filler = [''] * (len(cols) - 1)
-        for i, row in enumerate(table):
-            last = textwrap.wrap(row[-1], width=wrap_width)
-            print >> stream, format % tuple(row[:-1] + [last[0]])
-            for wrapped in last[1:]:
-                print >> stream, format % tuple(filler + [wrapped])
-            if vsep != None:
-                if i < len(table) - 1:
-                    print vsep * total_width
-    else:
-        total_width = sum(col_widths) + (len(sep) * (len(cols) - 1))
-        for i, row in enumerate(table):
-            print >> stream, format % tuple(row)
-            if vsep is not None:
-                if i < len(table) - 1:
-                    print vsep * total_width
-
-
 def _expand_period_contraints(periods):
     """
     period=B -> period like B%
@@ -154,6 +66,7 @@ def search_query(client,
                  project_name='Atlas_Production',
                  processing_step_name='Atlas_Production',
                  show_archived=False,
+                 literal_match=False,
                  **kwargs):
     try:
         table = TABLES[entity]
@@ -169,16 +82,40 @@ def search_query(client,
     if cmd_args is None:
         cmd_args = {}
 
-    if pattern is None:
-        pattern = '%'
+    if not isinstance(pattern, list):
+        patterns = [pattern]
     else:
-        pattern = re.sub('%+', '%', pattern)
-        if not pattern.startswith('%'):
-            pattern = '%' + pattern
-        if not pattern.endswith('%'):
-            pattern += '%'
+        patterns = pattern
+
+    constraints = []
+    for pattern in patterns:
+        # If the user has not put any '%' characters
+        # then we add them to the beginning and the end of the pattern
+        # otherwise assume the user knows what he/she is doing.
+        # If we do not do this it is impossible to search for strings which
+        # start with a given character sequence
+        if pattern is None:
+            if literal_match:
+                raise ValueError(
+                    'pattern must not be None for literal matches')
+            pattern = '%'
+
+        elif '%' not in pattern and not literal_match:
+            pattern = '%' + pattern + '%'
+
+        elif not literal_match:
+            # replace repeated % with a single %
+            pattern = re.sub('%+', '%', pattern)
+
+        if literal_match:
+            constraints.append("%s='%s'" % (primary_field, pattern))
+
+        else:
+            constraints.append("%s like '%s'" % (primary_field, pattern))
+
+    constraints = ' OR '.join(constraints)
+    constraints = '(%s)' % constraints
 
-    constraints = "%s like '%s'" % (primary_field, pattern)
     if kwargs:
         for name, value in kwargs.items():
             if value is not None:
@@ -223,6 +160,7 @@ def search_query(client,
         args.append('showArchived=true')
 
     result = client.execute(args)
+
     things = [thing for thing in result.rows()]
     if flatten:
         things = flatten_results(things, query_fields)
@@ -243,9 +181,10 @@ def get_types(client,
     """
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, types = search_query(client=client, entity='data_type', pattern=pattern,
-                                       processing_step_name='*',
-                                       order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, types = search_query(
+            client=client, entity='data_type', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         types = flatten_results(types, query_fields)
     return types
@@ -265,9 +204,10 @@ def get_subtypes(client,
     """
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, types = search_query(client=client, entity='subData_type', pattern=pattern,
-                                       processing_step_name='*',
-                                       order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, types = search_query(
+            client=client, entity='subData_type', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         types = flatten_results(types, query_fields)
     return types
@@ -294,9 +234,10 @@ def get_nomenclatures(client,
     """
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, nomens = search_query(client=client, entity='nomenclature', pattern=pattern,
-                                        processing_step_name='*',
-                                        order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, nomens = search_query(
+            client=client, entity='nomenclature', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         nomens = flatten_results(nomens, query_fields)
     return nomens
@@ -312,9 +253,10 @@ def get_projects(client,
                  **kwargs):
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, projects = search_query(client=client, entity='projects', pattern=pattern,
-                                          processing_step_name='*',
-                                          order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, projects = search_query(
+            client=client, entity='projects', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         projects = flatten_results(projects, query_fields)
     return projects
@@ -330,9 +272,10 @@ def get_subprojects(client,
                     **kwargs):
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, projects = search_query(client=client, entity='subProjects', pattern=pattern,
-                                          processing_step_name='*',
-                                          order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, projects = search_query(
+            client=client, entity='subProjects', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         projects = flatten_results(projects, query_fields)
     return projects
@@ -348,9 +291,10 @@ def get_prodsteps(client,
                   **kwargs):
     if 'write_status' not in kwargs:
         kwargs['write_status'] = 'valid'
-    query_fields, steps = search_query(client=client, entity='productionStep', pattern=pattern,
-                                       processing_step_name='*',
-                                       order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
+    query_fields, steps = search_query(
+            client=client, entity='productionStep', pattern=pattern,
+            processing_step_name='*',
+            order=order, limit=limit, fields=fields, show_archived=show_archived, **kwargs)
     if flatten:
         steps = flatten_results(steps, query_fields)
     return steps
@@ -364,6 +308,7 @@ def get_datasets(client,
                  fields=None,
                  flatten=False,
                  show_archived=False,
+                 from_file=False,
                  **kwargs):
     """
     Return list of datasets matching pattern
@@ -373,17 +318,27 @@ def get_datasets(client,
     cmd_args = {}
     if parent_type is not None and 'parent_type' not in kwargs:
         cmd_args['parentType'] = parent_type
-    pattern = __clean_dataset(pattern)
-    query_fields, datasets = search_query(client=client,
-                                          cmd='DatasetSearchQuery',
-                                          cmd_args=cmd_args,
-                                          entity='dataset',
-                                          pattern=pattern,
-                                          order=order, limit=limit,
-                                          fields=fields,
-                                          show_archived=show_archived, **kwargs)
+
+    if from_file:
+        patterns = read_patterns_from(pattern)
+    else:
+        patterns = [pattern]
+
+    patterns = [_clean_dataset(p) for p in patterns]
+
+    query_fields, datasets = search_query(
+            client=client,
+            cmd='DatasetSearchQuery',
+            cmd_args=cmd_args,
+            entity='dataset',
+            pattern=patterns,
+            order=order, limit=limit,
+            fields=fields,
+            show_archived=show_archived, **kwargs)
+
     if flatten:
         datasets = flatten_results(datasets, query_fields)
+
     return datasets
 
 
@@ -392,8 +347,14 @@ def get_periods_for_run(client, run):
     Return data periods which contain this run
     """
     result = client.execute(['GetDataPeriodsForRun', '-runNumber=%i' % run])
-    periods = sorted([RunPeriod(level=int(e['periodLevel']), name=str(e['period']), project=str(e['project'])) \
-                      for e in result.to_dict()['Element_Info'].values() ])
+    periods = sorted([
+        RunPeriod(
+            level=int(e['periodLevel']),
+            name=str(e['period']),
+            project=str(e['project']),
+            status=str(e['status']),
+            description=str(e['description']))
+        for e in result.to_dict()['Element_Info'].values()])
     return periods
 
 
@@ -491,7 +452,7 @@ def get_provenance(client, dataset, type=None, **kwargs):
     """
     Return all parent dataset of the given dataset
     """
-    dataset = __clean_dataset(dataset)
+    dataset = _clean_dataset(dataset)
     args = ["ListDatasetProvenance",
             "logicalDatasetName=%s" % dataset,
             'output=xml']
@@ -535,7 +496,7 @@ def get_dataset_info(client, dataset, **kwargs):
 
     *kwargs*: dict
     """
-    dataset = __clean_dataset(dataset)
+    dataset = _clean_dataset(dataset)
     args = ["GetDatasetInfo",
             "logicalDatasetName=%s" % dataset]
     if kwargs:
@@ -608,7 +569,7 @@ def get_event_info(client, dataset, **kwargs):
 
     *kwargs*: dict
     """
-    dataset = __clean_dataset(dataset)
+    dataset = _clean_dataset(dataset)
     if 'EVNT' not in dataset:
         prov = get_provenance(client, dataset, type='EVNT', **kwargs)
         evgen_datasets = []
@@ -693,8 +654,7 @@ def get_data_datasets(client,
                       fields=None,
                       latest=False,
                       flatten=False,
-                      **kwargs
-                     ):
+                      **kwargs):
     """
     *client*: AMIClient
 
@@ -722,9 +682,7 @@ def get_data_datasets(client,
     else list of tuples with elements in same order as fields
     """
 
-    """
-        Transmit period(s) as kwargs in order to do only one query
-    """
+    # Transmit period(s) as kwargs in order to do only one query
     if periods is not None:
         if isinstance(periods, basestring):
             periods = periods.split(',')
@@ -847,7 +805,7 @@ def get_files(client, dataset, limit=None):
 
     *limit*: [ tuple | list | int | None ]
     """
-    dataset = __clean_dataset(dataset)
+    dataset = _clean_dataset(dataset)
     args = ['ListFiles', 'logicalDatasetName=%s' % dataset]
     if limit is not None:
         if isinstance(limit, (list, tuple)):
@@ -859,52 +817,6 @@ def get_files(client, dataset, limit=None):
     return result.rows()
 
 
-def humanize_bytes(bytes, precision=1):
-    """
-    *bytes* : int
-        number of bytes
-
-    *precision* : int
-        precision at which to display human-readable form
-
-    Return a humanized string representation of a number of bytes.
-
-    Assumes `from __future__ import division`.
-
-    >>> from pyAMI.query import humanize_bytes
-    >>> humanize_bytes(1)
-    '1 byte'
-    >>> humanize_bytes(1024)
-    '1.0 kB'
-    >>> humanize_bytes(1024*123)
-    '123.0 kB'
-    >>> humanize_bytes(1024*12342)
-    '12.1 MB'
-    >>> humanize_bytes(1024*12342,2)
-    '12.05 MB'
-    >>> humanize_bytes(1024*1234,2)
-    '1.21 MB'
-    >>> humanize_bytes(1024*1234*1111,2)
-    '1.31 GB'
-    >>> humanize_bytes(1024*1234*1111,1)
-    '1.3 GB'
-    """
-    abbrevs = (
-        (1 << 50L, 'PB'),
-        (1 << 40L, 'TB'),
-        (1 << 30L, 'GB'),
-        (1 << 20L, 'MB'),
-        (1 << 10L, 'kB'),
-        (1, 'bytes')
-    )
-    if bytes == 1:
-        return '1 byte'
-    for factor, suffix in abbrevs:
-        if bytes >= factor:
-            break
-    return '%.*f %s' % (precision, bytes / factor, suffix)
-
-
 def list_files(client, dataset, limit=None, total=False, human_readable=False, long=False, stream=None):
     """
     *client*: AMIClient
@@ -945,18 +857,3 @@ def list_files(client, dataset, limit=None, total=False, human_readable=False, l
     else:
         for file in get_files(client, dataset, limit=limit):
             print >> stream, file['LFN']
-
-
-def print_dict(d, sep=' ', stream=None):
-    """
-    *d* : dict
-        dictionary to print
-
-    *stream* : file
-        file-like object to print output on
-        defaults to sys.stdout
-    """
-    if stream is None:
-        stream = sys.stdout
-    table = [[name, ':', value] for name, value in d.items()]
-    print_table(table, sep=' ', stream=stream)
